# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from goodtables import validate


# Validate

def test_check_blacklisted_value(log):
    source = [
        ['row', 'name'],
        [2, 'Alex'],
        [3, 'John'],
        [4, 'mistake'],
        [5, 'error'],
        [6],
    ]
    report = validate(source, checks=[
        {'blacklisted-value': {'column': 1, 'blacklist': [10]}},
        {'blacklisted-value': {'column': 2, 'blacklist': ['mistake']}},
        {'blacklisted-value': {'column': 'row', 'blacklist': [10]}},
        {'blacklisted-value': {'column': 'name', 'blacklist': ['error']}},
    ])
    assert log(report) == [
        (1, 4, 2, 'blacklisted-value'),
        (1, 5, 2, 'blacklisted-value'),
    ]


def test_check_blacklisted_value_non_existent_column(log):
    source = [
        ['row', 'name'],
        [2, 'Alex'],
    ]
    report = validate(source, checks=[
        {'blacklisted-value': {'column': 3, 'blacklist': [10]}},
        {'blacklisted-value': {'column': 'non-existent', 'blacklist': ['mistake']}},
    ])
    assert log(report) == [
        (1, 2, None, 'blacklisted-value'),
        (1, 2, None, 'blacklisted-value'),
    ]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..registry import check
from ..error import Error


# Module API

@check('unique-constraint')
class UniqueConstraint(object):

    # Public

    def __init__(self, **options):
        self.__unique_fields_cache = None

    def check_row(self, cells):
        errors = []

        # Prepare unique checks
        if self.__unique_fields_cache is None:
            self.__unique_fields_cache = _create_unique_fields_cache(cells)

        # Check unique
        for column_numbers, cache in self.__unique_fields_cache.items():
            column_cells = tuple(
                cell
                for column_number, cell in enumerate(cells, start=1)
                if column_number in column_numbers
            )
            column_values = tuple(cell.get('value') for cell in column_cells)
            row_number = column_cells[0]['row-number']

            all_values_are_none = (set(column_values) == {None})
            if not all_values_are_none:
                if column_values in cache['data']:
                    message_substitutions = {
                        'row_numbers': ', '.join(map(str, cache['refs'] + [row_number])),
                    }

                    # FIXME: The unique constraint can be related to multiple
                    # columns (e.g. a composite primary key), but here we only
                    # pass the 1st column.
                    error = Error(
                        'unique-constraint',
                        column_cells[0],
                        message_substitutions=message_substitutions
                    )
                    errors.append(error)
                cache['data'].add(column_values)
                cache['refs'].append(row_number)

        return errors

# Internal


def _create_unique_fields_cache(cells):
    primary_key_column_numbers = []
    cache = {}

    # Unique
    for column_number, cell in enumerate(cells, start=1):
        field = cell.get('field')
        if field is not None:
            if field.descriptor.get('primaryKey'):
                primary_key_column_numbers.append(column_number)
            if field.constraints.get('unique'):
                cache[tuple([column_number])] = {
                    'data': set(),
                    'refs': [],
                }

    # Primary key
    if primary_key_column_numbers:
        cache[tuple(primary_key_column_numbers)] = {
            'data': set(),
            'refs': [],
        }

    return cache



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import click
from click_default_group import DefaultGroup
import goodtables
import json as json_module
from pprint import pformat
click.disable_unicode_literals_warning = True


# Module API

@click.group(cls=DefaultGroup, default='validate', default_if_no_args=True)
@click.version_option(goodtables.__version__, message='%(version)s')
def cli():
    """Tabular files validator.

    There are two categories of validation checks available:

    * Structural checks: ensure there are no empty rows, no blank headers, etc.

    * Content checks: ensure the values have the correct types (e.g. string),
      their format is valid (e.g. e-mail), and they respect some constraint
      (e.g. age is greater than 18).

    \b
    Full documentation at: <https://github.com/frictionlessdata/goodtables-py/>
    """
    pass


@cli.command(
    short_help='Validate tabular files (default).',
)
@click.argument('paths', type=click.Path(), nargs=-1, required=True)
@click.option('--quiet', '-q', is_flag=True, help='Don\'t output anything.')
@click.option('--json', is_flag=True, help='Output report as JSON.')
@click.option(
    '--output',
    '-o',
    type=click.File('w'),
    default='-',
    help='Redirect output to a file.'
)
@click.option('--preset')
@click.option('--schema', type=click.Path(), help='Path to a Table Schema.')
@click.option(
    '--infer-schema/--no-infer-schema',
    default=False,
    help='Infer schema. If an explicit schema is defined, infer missing columns only.'
)
@click.option('--checks', '-c', multiple=True, help='Checks to enable.')
@click.option(
    '--skip-checks',
    '-C',
    multiple=True,
    help='Checks to disable.'
)
@click.option(
    '--order-fields',
    is_flag=True,
    help='Don\'t validate the columns order.'
)
@click.option(
    '--row-limit',
    type=int,
    default=-1,
    help='Maximum number of rows to validate (-1 for no limit)'
)
@click.option(
    '--table-limit',
    type=int,
    default=-1,
    help='Maximum number of tables to validate (-1 for no limit)'
)
@click.option(
    '--error-limit',
    type=int,
    default=-1,
    help='Stop validating if there are more than this number of errors (-1 for no limit).'
)
def validate(paths, json, **options):
    # Remove blank values
    options = {key: value for key, value in options.items() if value is not None}
    if not options['checks']:
        del options['checks']
    if not options['skip_checks']:
        del options['skip_checks']

    options['infer_fields'] = options['infer_schema']
    quiet = options.pop('quiet')
    output = options.pop('output')

    sources = [{'source': path} for path in paths]
    schema = options.pop('schema', None)
    if schema:
        for source in sources:
            source['schema'] = schema

    report = goodtables.validate(sources, **options)

    if not quiet:
        _print_report(report, output=output, json=json)

    exit(not report['valid'])


@cli.command()
@click.argument('paths', type=click.Path(), nargs=-1, required=True)
@click.option(
    '--output',
    '-o',
    type=click.File('w'),
    default='-',
    help='Redirect output to a file.'
)
def init(paths, output, **kwargs):
    """Init data package from list of files.

    It will also infer tabular data's schemas from their contents.
    """
    dp = goodtables.init_datapackage(paths)

    click.secho(
        json_module.dumps(dp.descriptor, indent=4),
        file=output
    )

    exit(dp.valid)  # Just to be defensive, as it should always be valid.


# Internal
def _print_report(report, output=None, json=False):
    def secho(*args, **kwargs):
        click.secho(file=output, *args, **kwargs)

    if json:
        return secho(json_module.dumps(report, indent=4))

    color = 'green' if report['valid'] else 'red'
    tables = report.pop('tables')
    warnings = report.pop('warnings')
    secho('DATASET', bold=True)
    secho('='*7, bold=True)
    secho(pformat(report), fg=color, bold=True)
    if warnings:
        secho('-'*9, bold=True)
    for warning in warnings:
        secho('Warning: %s' % warning, fg='yellow')
    for table_number, table in enumerate(tables, start=1):
        secho('\nTABLE [%s]' % table_number, bold=True)
        secho('='*9, bold=True)
        color = 'green' if table['valid'] else 'red'
        errors = table.pop('errors')
        secho(pformat(table), fg=color, bold=True)
        if errors:
            secho('-'*9, bold=True)
        for error in errors:
            template = '[{row-number},{column-number}] [{code}] {message}'
            substitutions = {
                'row-number': error.get('row-number', '-'),
                'column-number': error.get('column-number', '-'),
                'code': error.get('code', '-'),
                'message': error.get('message', '-'),
            }
            message = template.format(**substitutions)
            secho(message)


# Main program

if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ...registry import check
from ...error import Error


# Module API

@check('blacklisted-value', type='custom', context='body')
class BlacklistedValue(object):

    # Public

    def __init__(self, column, blacklist, **options):
        self.__column = column
        self.__blacklist = blacklist

    def check_row(self, cells):

        # Get cell
        cell = None
        for item in cells:
            if self.__column in [item['column-number'], item['header']]:
                cell = item
                break

        # Check cell
        if not cell:
            message = 'Blacklisted value check requires column "{column}" to exist'
            message = message.format(column=self.__column)
            error = Error(
                'blacklisted-value',
                row_number=cells[0]['row-number'],
                message=message
            )
            return [error]

        # Check value
        value = cell.get('value')
        if value in self.__blacklist:
            message = (
                'Value "{value}" in column {column_number} on row {row_number}'
                ' is blacklisted'
            )
            message_substitutions = {
                'value': value,
            }
            error = Error(
                'blacklisted-value',
                cell,
                message=message,
                message_substitutions=message_substitutions
            )
            return [error]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals


# General

DEFAULT_ERROR_LIMIT = 1000
DEFAULT_TABLE_LIMIT = 10
DEFAULT_ROW_LIMIT = 1000

# Presets

PRESETS = [
    'goodtables.presets.table',
    'goodtables.presets.datapackage',
    'goodtables.presets.nested',
]

# Checks

CHECKS = [
    # Spec
    'goodtables.checks.blank_header',
    'goodtables.checks.duplicate_header',
    'goodtables.checks.non_matching_header',
    'goodtables.checks.extra_header',
    'goodtables.checks.missing_header',
    'goodtables.checks.blank_row',
    'goodtables.checks.duplicate_row',
    'goodtables.checks.extra_value',
    'goodtables.checks.missing_value',
    'goodtables.checks.type_or_format_error',
    'goodtables.checks.required_constraint',
    'goodtables.checks.pattern_constraint',
    'goodtables.checks.unique_constraint',
    'goodtables.checks.enumerable_constraint',
    'goodtables.checks.minimum_constraint',
    'goodtables.checks.maximum_constraint',
    'goodtables.checks.minimum_length_constraint',
    'goodtables.checks.maximum_length_constraint',
    # Contrib
    'goodtables.contrib.checks.blacklisted_value',
    'goodtables.contrib.checks.deviated_value',
    'goodtables.contrib.checks.foreign_key',
    'goodtables.contrib.checks.sequential_value',
    'goodtables.contrib.checks.truncated_value',
    'goodtables.contrib.checks.custom_constraint',
]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import json
from copy import deepcopy
from collections import defaultdict
from datapackage import Package
from ...registry import check
from ...error import Error


# Module API

@check('foreign-key', type='custom', context='body')
class ForeignKey(object):

    # Public

    def __init__(self, **options):
        self.__package = None
        self.__schema = None
        self.__relations = None
        self.__code = 'foreign-key'

    def prepare(self, stream, schema, extra):

        # Prepare package
        if 'datapackage' not in extra or 'resource-name' not in extra:
            return False
        descriptor = extra['datapackage']
        if descriptor.strip().startswith('{'):
            descriptor = json.loads(descriptor)
        self.__package = Package(descriptor)

        # Prepare schema
        if not schema:
            return False
        if not schema.foreign_keys:
            return False
        self.__schema = schema

        # Prepare foreign keys values
        relations = _get_relations(
            self.__package, self.__schema,
            current_resource_name=extra['resource-name'])
        self.__foreign_keys_values = _get_foreign_keys_values(
            self.__schema, relations)

        return True

    def check_row(self, cells):
        row_number = cells[0]['row-number']

        # Prepare keyed_row
        keyed_row = {}
        for cell in cells:
            if cell.get('field'):
                keyed_row[cell.get('field').name] = cell.get('value')

        # Resolve relations
        errors = []
        for foreign_key in self.__schema.foreign_keys:
            success = _resolve_relations(
                deepcopy(keyed_row), self.__foreign_keys_values, foreign_key)
            if success is None:
                message = 'Foreign key "{fields}" violation in row {row_number}'
                message_substitutions = {'fields': foreign_key['fields']}
                errors.append(Error(
                    self.__code,
                    row_number=row_number,
                    message=message,
                    message_substitutions=message_substitutions,
                ))

        return errors


# Internal

def _get_relations(package, schema, current_resource_name=None):
    # It's based on the following code:
    # https://github.com/frictionlessdata/datapackage-py/blob/master/datapackage/resource.py#L393

    # Prepare relations
    relations = {}
    for fk in schema.foreign_keys:
        resource_name = fk['reference'].get('resource')
        package_name = fk['reference'].get('package')
        resource = None

        # Self-referenced resource
        if not resource_name:
            for item in package.resources:
                if item.name == current_resource_name:
                    resource = item

        # Internal resource
        elif not package_name:
            resource = package.get_resource(resource_name)

        # External resource (experimental)
        # For now, we rely on uniqueness of resource names and support relative paths
        else:
            descriptor = package_name
            if not descriptor.startswith('http'):
                descriptor = '/'.join([package.base_path, package_name])
            package = Package(descriptor)
            resource = package.get_resource(resource_name)

        # Add to relations (can be None)
        relations[resource_name] = resource
        if resource and resource.tabular:
            relations[resource_name] = resource.read(keyed=True)

    return relations


def _get_foreign_keys_values(schema, relations):
    # It's based on the following code:
    # https://github.com/frictionlessdata/tableschema-py/blob/master/tableschema/table.py#L218

    # we dont need to load the complete reference table to test relations
    # we can lower payload AND optimize testing foreign keys
    # by preparing the right index based on the foreign key definition
    # foreign_keys are sets of tuples of all possible values in the foreign table
    # foreign keys =
    # [reference] [foreign_keys tuple] = { (foreign_keys_values, ) : one_keyedrow, ... }
    foreign_keys = defaultdict(dict)
    if schema:
        for fk in schema.foreign_keys:
            # load relation data
            relation = fk['reference']['resource']

            # create a set of foreign keys
            # to optimize we prepare index of existing values
            # this index should use reference + foreign_keys as key
            # cause many foreign keys may use the same reference
            foreign_keys[relation][tuple(fk['reference']['fields'])] = {}
            for row in (relations[relation] or []):
                key = tuple([row[foreign_field] for foreign_field in fk['reference']['fields']])
                # here we should chose to pick the first or nth row which match
                # previous implementation picked the first, so be it
                if key not in foreign_keys[relation][tuple(fk['reference']['fields'])]:
                    foreign_keys[relation][tuple(fk['reference']['fields'])][key] = row
    return foreign_keys


def _resolve_relations(keyed_row, foreign_keys_values, foreign_key):
    # It's based on the following code:
    # https://github.com/frictionlessdata/tableschema-py/blob/master/tableschema/table.py#L228

    # local values of the FK
    local_values = tuple(keyed_row[f] for f in foreign_key['fields'])
    if len([l for l in local_values if l]) > 0:
        # test existence into the foreign
        relation = foreign_key['reference']['resource']
        keys = tuple(foreign_key['reference']['fields'])
        foreign_values = foreign_keys_values[relation][keys]
        if local_values in foreign_values:
            return foreign_values[local_values]
        else:
            return None
    else:
        # empty values for all keys, return original values
        return keyed_row



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import copy
from tableschema import Schema
from ..registry import check
from ..error import Error


# Module API

@check('extra-header')
class ExtraHeader(object):

    # Public
    def __init__(self, infer_fields=False, **options):
        self.__infer_fields = infer_fields

    def check_headers(self, cells, sample):
        errors = []

        for cell in copy(cells):

            # Skip if cell has field
            if 'field' in cell and cell['field'] is not None:
                continue

            # Infer field
            if self.__infer_fields:
                column_sample = []
                for row in sample:
                    value = None
                    if len(row) >= cell['column-number']:
                        value = row[cell['column-number'] - 1]
                    column_sample.append([value])
                schema = Schema()
                schema.infer(column_sample, headers=[cell.get('header')])
                cell['field'] = schema.fields[0]

            # Add error/remove column
            else:
                message_substitutions = {
                    'header': '"{}"'.format(cell.get('header')),
                }
                error = Error(
                    'extra-header',
                    cell,
                    message_substitutions=message_substitutions
                )
                errors.append(error)
                cells.remove(cell)

        return errors



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import sys
import pytest
from goodtables import validate


# Validate

@pytest.mark.skipif(sys.version_info >= (3, 8), reason='Problem with Python3.8+')
def test_check_custom_constraint(log):
    source = [
        ['row', 'salary', 'bonus'],
        [2, 1000, 200],
        [3, 2500, 500],
        [4, 1300, 500],
        [5, 5000, 1000],
        [6],
    ]
    report = validate(source, checks=[
        {'custom-constraint': {'constraint': 'salary > bonus * 4'}},
    ])
    assert log(report) == [
        (1, 4, None, 'custom-constraint'),
        (1, 6, None, 'custom-constraint'),
    ]


def test_check_custom_constraint_incorrect_constraint(log):
    source = [
        ['row', 'name'],
        [2, 'Alex'],
    ]
    report = validate(source, checks=[
        {'custom-constraint': {'constraint': 'vars()'}},
        {'custom-constraint': {'constraint': 'import(os)'}},
        {'custom-constraint': {'constraint': 'non_existent > 0'}},
    ])
    assert log(report) == [
        (1, 2, None, 'custom-constraint'),
        (1, 2, None, 'custom-constraint'),
        (1, 2, None, 'custom-constraint'),
    ]
